from __future__ import print_function

import ast
import os
from collections import namedtuple

import mock
import pytest
from flake8.options.manager import OptionManager

from flake8_illegal_import import (Flake8Argparse, ImportChecker,
                                   format_code, resolve_path)

curr_dir_path = os.path.dirname(os.path.realpath(__file__))
SAMPLE_FILE_PATH = curr_dir_path + '/tests/files/sample.py'
SAMPLE_FILE_DIR = os.path.dirname(SAMPLE_FILE_PATH)

SysArgs = namedtuple('SysArgs', 'illegal_import_dir illegal_import_packages')


def get_tree(filename):
    with open(filename, 'rb') as f:
        return compile(f.read(), filename, 'exec', ast.PyCF_ONLY_AST, True)


def get_tree_from_str(pycode, filename="test.py"):
    return compile(pycode, filename, 'exec', ast.PyCF_ONLY_AST, True)


class TestModuleUtils():
    def test__resolve_path_absolute(self):
        assert resolve_path('/tmp') == '/tmp'

    def test__resolve_path_relative(self):
        assert resolve_path('./tmp') == os.path.abspath('./tmp')

    def test__resolve_path_expand(self):
        assert resolve_path('~/tmp') == os.path.expanduser('~/tmp')

    def test__format_code(self):
        assert format_code(302) == "II302"


class TestFlake8Optparse():
    def test__add_options(self):
        flake8_opt_mgr = OptionManager()
        plugin = Flake8Argparse(None, SAMPLE_FILE_PATH)
        plugin.add_options(flake8_opt_mgr)
        assert len(flake8_opt_mgr.options) == 2

    def test__parse_options(self):
        flake8_opt_mgr = OptionManager()
        plugin = Flake8Argparse(None, SAMPLE_FILE_PATH)
        args = SysArgs(illegal_import_dir=SAMPLE_FILE_DIR, illegal_import_packages='os,json')
        plugin.parse_options(flake8_opt_mgr, args, extra_args=None)
        assert plugin.illegal_import_dir == args.illegal_import_dir
        assert plugin.illegal_import_packages == ['os', 'json']


class TestImportChecker():

    @pytest.mark.parametrize("code_snippet", [
        "",
        "import os\n",
        "import os as forbidden\n",
        "import os # forbidden\n",
    ])
    def test__get_illegal_imports__no_match(self, code_snippet):
        tree = get_tree_from_str(code_snippet)

        banned_packages = ['forbidden']
        illegals_imports = ImportChecker.get_illegal_imports(tree, banned_packages)
        errors = {node.lineno: pkg_name for node, pkg_name in illegals_imports}

        assert errors == {}

    @pytest.mark.parametrize("code_snippet,expected_errors", [
        (
                "import forbidden\n",
                {1: 'forbidden'}
        ),
        (
                "import forbidden\nimport forbidden\n",
                {1: 'forbidden',
                 2: 'forbidden'}
        ),
        (
                "import forbidden as forbidden\n",
                {1: 'forbidden'}
        ),
        (
                "import forbidden as os\n",
                {1: 'forbidden'}
        ),
        (
                "import os;\nimport json;import forbidden\n",
                {2: 'forbidden'}
        ),
    ])
    def test__get_illegal_imports__import_simple_case(self, code_snippet, expected_errors):
        tree = get_tree_from_str(code_snippet)

        banned_packages = ['forbidden']
        illegals_imports = ImportChecker.get_illegal_imports(tree, banned_packages)
        errors = {node.lineno: pkg_name for node, pkg_name in illegals_imports}

        assert errors == expected_errors

    @pytest.mark.parametrize("code_snippet,expected_errors", [
        (
                "from forbidden import really_forbidden\n",
                {1: 'forbidden'}
        ),
        (
                "from forbidden import really_forbidden as os\n",
                {1: 'forbidden'}
        ),
        (
                "import json, os\nfrom forbidden import really_forbidden",
                {2: 'forbidden'}
        ),
    ])
    def test__get_illegal_imports__importfrom_simple_case(self, code_snippet, expected_errors):
        tree = get_tree_from_str(code_snippet)

        banned_packages = ['forbidden']
        illegals_imports = ImportChecker.get_illegal_imports(tree, banned_packages)
        errors = {node.lineno: pkg_name for node, pkg_name in illegals_imports}

        assert errors == expected_errors

    def test__get_illegal_imports__coma_separated(self):
        code = "import json, forbidden"
        tree = get_tree_from_str(code)

        banned_packages = ['forbidden']
        illegals_imports = ImportChecker.get_illegal_imports(tree, banned_packages)
        errors = {node.lineno: pkg_name for node, pkg_name in illegals_imports}

        assert errors == {1: 'forbidden'}

    def test__run(self):
        tree = get_tree(SAMPLE_FILE_PATH)
        args = SysArgs(illegal_import_dir='./', illegal_import_packages='os,json,abc')

        checker = ImportChecker(tree, SAMPLE_FILE_PATH)
        checker.parse_options(None, args, None)

        expected = [
            (1, 0, 'II101 importing this package is forbidden in this directory (os)', ImportChecker),
            (4, 0, 'II101 importing this package is forbidden in this directory (json)', ImportChecker),
            (5, 0, 'II101 importing this package is forbidden in this directory (abc)', ImportChecker),
            (7, 0, 'II101 importing this package is forbidden in this directory (os)', ImportChecker),
            (8, 0, 'II101 importing this package is forbidden in this directory (os)', ImportChecker),
            (14, 4, 'II101 importing this package is forbidden in this directory (os)', ImportChecker),
            (15, 4, 'II101 importing this package is forbidden in this directory (json)', ImportChecker),
        ]
        res = list(checker.run())
        assert res == expected

    @mock.patch('flake8_illegal_import.ImportChecker.report')
    def test__run__no_package_name(self, reporter):
        tree = get_tree(SAMPLE_FILE_PATH)
        args = SysArgs(illegal_import_dir='/tmp', illegal_import_packages='')

        checker = ImportChecker(tree, SAMPLE_FILE_PATH)
        checker.parse_options(None, args, None)
        resp = list(checker.run())
        assert len(resp) == 0
        reporter.assert_called_once_with('No illegal import package set - skip checks')

    @mock.patch('flake8_illegal_import.ImportChecker.report')
    def test__run__dir_not_exist(self, reporter):
        tree = get_tree(SAMPLE_FILE_PATH)
        args = SysArgs(illegal_import_dir='./non-exist', illegal_import_packages='os')

        checker = ImportChecker(tree, SAMPLE_FILE_PATH)
        checker.parse_options(None, args, None)
        resp = list(checker.run())
        assert len(resp) == 0
        reporter.assert_called_once()
        assert 'WARNING: Directory configured does not exist:' in reporter.call_args_list[0][0][0]



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Extension for flake8 to reject specific imports"""
from __future__ import print_function, unicode_literals

import ast
from collections import namedtuple
import os.path

__version__ = '0.1.0'


CODE_PREFIX = "II"


def format_code(code):
    return '{}{}'.format(CODE_PREFIX, code)


def resolve_path(dir_path):
    dir_path = os.path.expanduser(dir_path)
    return os.path.abspath(dir_path)


class Flake8Argparse(object):

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = resolve_path(filename)

    @classmethod
    def add_options(cls, option_manager):
        option_manager.add_option(
            "--illegal-import-dir", type="str", default="./",
            parse_from_config=True,
            help="Directory to consider when looking for illegal import",
        )

        option_manager.add_option(
            "--illegal-import-packages", type="str", parse_from_config=True,
            help="Set packages that are not allowed in directory",
        )

    @classmethod
    def parse_options(cls, option_manager, options, extra_args):
        cls.illegal_import_dir = resolve_path(options.illegal_import_dir)
        illegal_import_packages = options.illegal_import_packages or ""
        cls.illegal_import_packages = [pkg for pkg in illegal_import_packages.split(',') if pkg]    # Allows for "package," as option


def root_package_name(name):
    tree = ast.parse(name)
    for node in ast.walk(tree):
        if isinstance(node, ast.Name):
            return node.id
    else:
        return None


ImportedPackage = namedtuple('ImportedPackage', 'name node')


class ImportVisitor(ast.NodeVisitor):

    def __init__(self):
        self.imported_packages = []

    def visit_Import(self, node):  # noqa: N802
        modules = [alias.name for alias in node.names]
        for module in modules:
            self.imported_packages.append(
                ImportedPackage(root_package_name(module), node)
            )

    def visit_ImportFrom(self, node):  # noqa: N802
        module = node.module or ''
        if node.level > 0:
            # ignore application relative
            return

        self.imported_packages.append(
            ImportedPackage(root_package_name(module), node)
        )


class ImportChecker(Flake8Argparse):

    version = __version__
    name = 'illegal-import'

    ERRORS = {
        101: 'importing this package is forbidden in this directory ({package})',
    }

    def _generate_error(self, node, code, **params):
        msg = '{0} {1}'.format(format_code(code), self.ERRORS[code])
        msg = msg.format(**params)
        return node.lineno, node.col_offset, msg, type(self)

    @staticmethod
    def get_illegal_imports(tree, banned_packages):
        visitor = ImportVisitor()
        visitor.visit(tree)

        for package_name, node in visitor.imported_packages:
            if package_name in banned_packages:
                yield (node, package_name)

    def report(self, msg):
        print(msg)

    def run(self):
        dir_path = self.illegal_import_dir
        banned_packages = set(self.illegal_import_packages)

        if not banned_packages:
            self.report('No illegal import package set - skip checks')
            return

        if not os.path.isdir(dir_path):
            self.report('WARNING: Directory configured does not exist: {}'.format(dir_path))
            return

        if not self.filename.startswith(dir_path):
            # File not concerned by the restriction
            return

        for node, pkg in ImportChecker.get_illegal_imports(self.tree, banned_packages):
            yield self._generate_error(node, 101, package=pkg)



import os
import ast
import codecs
import json as jayson
import collections, abc

import os.path
from os.path import expanduser

from .test import nothing


def funky_method_that_import():
    from os.path import abspath
    from json import dumps
    return



# -*- coding: utf-8 -*-
import codecs

from setuptools import setup


def get_version():
    with open('flake8_illegal_import.py') as f:
        for line in f:
            if line.startswith('__version__'):
                return eval(line.split('=')[-1])


def get_prefix():
    with open('flake8_illegal_import.py') as f:
        for line in f:
            if line.startswith('CODE_PREFIX'):
                return eval(line.split('=')[-1])


def get_long_description():
    with codecs.open('README.rst', 'r', 'utf-8') as f:
        return f.read()


INSTALL_REQUIRES = ['flake8']
TESTS_REQUIRES = ['pytest', 'mock', 'pytest-cov']

setup(
    name='flake8-illegal-import',
    version=get_version(),
    description='illegal import detector, plugin for flake8',
    long_description=get_long_description(),
    keywords='flake8 import reject',
    maintainer='Bastien Gérard',
    maintainer_email='bastien.gerard@gmail.com',
    url='https://github.com/bagerard/flake8-illegal-import',
    license='MIT License',
    py_modules=['flake8_illegal_import'],
    zip_safe=False,
    entry_points={
        'flake8.extension': [
            '{prefix} = flake8_illegal_import:ImportChecker'.format(prefix=get_prefix()),
        ],
    },
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRES,
    setup_requires=['pytest-runner'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
    ],
)
